# -*- coding: utf8 -*-
from __future__ import unicode_literals

import errno
import select
import socket
import time
import traceback

from driverng.transport.base import (
    BaseTransport,
    TransportException,
    TransportTimedOutException,
)
from driverng.utils import byte2log


class TCPIHETransport(BaseTransport):
    def __init__(
        self,
        host_send=None,
        host_recv=None,
        port_send=None,
        port_recv=None,
        mode="mixte",
        **kwargs
    ):
        self.host_send = host_send
        super(TCPIHETransport, self).__init__(**kwargs)
        self.transport_send = None
        self.transport_recv = None
        if port_send is None:
            host_send, port_send = host_send.rsplit(":", 1)
            port_send = int(port_send)
        if port_recv is None:
            host_recv, port_recv = host_recv.rsplit(":", 1)
            port_recv = int(port_recv)
        self.host_recv = host_recv
        self.port_recv = port_recv
        self.host_send = host_send
        self.port_send = port_send

        if mode == "mixte":
            # Ouverture de deux sockets
            # Le canal de réception (entrant) est serveur TCP/IP
            s2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.set_keepalive(s2)
            try:
                s2.bind((host_recv, port_recv))
            except socket.error as e:
                if e.args[0] in (errno.EADDRINUSE,):
                    s2.close()
                    time.sleep(120)
                    raise TransportException(
                        "Port tcp receiver (errno=%d) Address already in use"
                        % (e.args[0],)
                    )
                else:
                    raise
            s2.listen(1)
            self.socket_recv = s2
            self.get_server_connexion()

            # Le canal d'émission (sortant) est client TCP/IP
            if host_send == "0.0.0.0":
                self.logger.info(
                    "Le port tcp sender est paramétré en serveur tcp/ip, vérifier l'adresse ip %s",
                    host_send,
                )
            else:
                self.get_client_connexion()
        self.transport = self.transport_send  # pour les tests

    def get_client_connexion(self):
        self.logger.info(
            "Port tcp sender: Try to connect to %s:%d", self.host_send, self.port_send
        )
        self.transport_send = None
        try:
            self.transport_send = socket.create_connection(
                (self.host_send, self.port_send), 10
            )
            self.set_keepalive(self.transport_send)
        except socket.gaierror as e:
            raise TransportException("Port tcp sender disconnected %s" % e)
        except socket.timeout:
            raise TransportException("Timeout")
        except socket.error as e:
            if e.args[0] in (errno.ECONNREFUSED,):
                raise TransportException(
                    "Port tcp sender disconnected (errno=%d)" % (e.args[0],)
                )
            else:
                raise
        else:
            self.logger.info(
                "Port sender connected to %s:%d", self.host_send, self.port_send
            )
        return True

    def get_server_connexion(self):
        self.logger.info(
            "Waiting for connection receiver on %s:%s", self.host_recv, self.port_recv
        )
        try:
            self.socket_recv.setblocking(False)
            self.socket_recv.settimeout(2)
            conn, addr = self.socket_recv.accept()
            self.logger.info("Port receiver connected by %s", addr)
        except (ValueError, socket.timeout):
            conn, addr = (None, None)
        self.transport_recv = conn

    @staticmethod
    def set_keepalive(s):
        s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        # Activate after 5 second of idle
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 5)

        # ping every 60 second
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 60)

        # Timeout after 2 failed ping
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 2)

    def fileno_send(self):
        return self.transport_send.fileno()

    def fileno_recv(self):
        return self.transport_recv.fileno()

    def set_timeout_port_sender(self, timeout):
        # Timeout sur le port d'émission (réservé au message HL7-IHE de type
        # LAB-28
        self.transport_send.settimeout(timeout)

    def set_timeout(self, timeout):
        # Timeout sur le port de réception (réservé aux messages HL7-IHE de
        # type LAB-27 et LAB-29
        self.transport_recv.settimeout(timeout)

    def poll(self, timeout, wqueue):
        rfds_send = [self.transport_send]
        if self.transport_recv is None:  # accepte une nouvelle connexion si non active
            self.get_server_connexion()
        rfds_recv = [self.transport_recv]

        if self.socket_send is not None:  # serveur
            rfds_send.append(self.socket_send)
        if self.socket_recv is not None:
            rfds_recv.append(self.socket_recv)

        r_send, _, _ = select.select(rfds_send, [], [], timeout)
        r_recv = []
        if self.socket_recv is not None and self.transport_recv is not None:
            r_recv, _, _ = select.select(rfds_recv, [], [], timeout)

        if self.socket_send is not None and self.socket_send in r_send:
            conn, addr = self.socket_send.accept()
            self.transport_send = conn
            self.logger.info("Connection Sender remplaced by %s", addr)
            return False

        if self.socket_recv is not None and self.socket_recv in r_recv:
            conn, addr = self.socket_recv.accept()
            self.transport_recv = conn
            self.logger.info("Connection Receiver remplaced by %s", addr)
            return False

        if self.transport_recv is not None and self.transport_recv in r_recv:
            return True
        else:
            return False
        self.transport = self.transport_send  # pour les tests

    def readline_ack(self, break_func=lambda buf: buf[-2:] == b"\r\n"):
        """Lecture d'un bloc de données sur le port d'émission, on attend
        l'acquittement du message de type LAB-27 (réservé au port d'émission)
        @param break_func: Fonction appelée à chaque bloc lu pour savoir
        s'il faut continuer de lire ou non. Par défaut c'est les caractères
        <cr><lf> en fin de buffer (protocole ASTM)
        """

        data = b""
        while not break_func(data):
            try:
                data += self._read_port_sender(4096)
            except Exception:
                if data:
                    self.logger.debug(
                        "Port d'émission: (before error) recv %s",
                        byte2log(data, truncate=True),
                    )
                raise

        self.logger.info("Lecture de l'acquittement sur le port d'émission")
        self.logger.info("recv %s", byte2log(data, truncate=True))
        return data

    def _read_port_sender(self, size):
        data = None

        r, _, _ = select.select(
            [self.fileno_send()], [], [], self.transport_send.gettimeout()
        )

        if not r:
            raise TransportTimedOutException("Timeout sur le port d'émission")

        try:
            data = self.transport_send.recv(size)
        except socket.error as e:
            if e.args[0] in (
                errno.EBADF,
                errno.ECONNRESET,
                errno.ENOTCONN,
                errno.ESHUTDOWN,
                errno.ECONNABORTED,
                errno.ETIMEDOUT,
            ):
                self.close()
                time.sleep(60)
                raise TransportException(
                    "Port d'émission - Disconnected (errno=%d)" % e.args[0]
                )
            else:
                self.close()
                time.sleep(60)
                raise

        if not data:
            self.close()
            time.sleep(60)
            raise TransportException("Port d'émission - Disconnected (recv 0 bytes)")

        return data

    def _read(self, size):
        data = None

        r, _, _ = select.select(
            [self.fileno_recv()], [], [], self.transport_recv.gettimeout()
        )

        if not r:
            raise TransportTimedOutException("Timeout sur le port de réception")

        try:
            data = self.transport_recv.recv(size)
        except socket.error as e:
            if e.args[0] in (
                errno.EBADF,
                errno.ECONNRESET,
                errno.ENOTCONN,
                errno.ESHUTDOWN,
                errno.ECONNABORTED,
            ):
                self.close()
                time.sleep(60)
                raise TransportException(
                    "Port de réception - Disconnected (errno=%d)" % e.args[0]
                )
            elif e.args[0] == errno.ETIMEDOUT:
                self.logger.info(
                    "Port de réception - Erreur timeout - Disconnected (errno=%d)"
                    % e.args[0]
                )
                self.poll(2, None)
                self.logger.info("Port de réception: Réinitialisation du port")
                return data if data is not None else ""
            else:
                self.logger.info("Port de réception - Erreur (errno=%d)" % e.args[0])
                self.poll(2, None)
                self.logger.info("Port de réception: Réinitialisation du port")
                return ""

        if not data:
            self.logger.info("Port de réception: Réinitialisation du port")
            time.sleep(60)
            self.poll(2, None)
            return b""
        return data

    def _write(self, data):

        channel_receiver = False
        if b"RSP^K11^RSP_K11" in data or b"ACK^R22^ACK" in data:
            channel_receiver = True
        if channel_receiver:
            # Réponse sur le canal de réception pour les acquittements
            _, w, _ = select.select(
                [], [self.fileno_recv()], [], self.transport_recv.gettimeout()
            )
        else:
            # Envoi des demandes sur le canal d'émission
            _, w, _ = select.select(
                [], [self.fileno_send()], [], self.transport_send.gettimeout()
            )

        if channel_receiver and not w:
            self.close()
            time.sleep(60)
            raise TransportException(
                "Port de réception: lors de l'envoi de l'acquittement - Erreur de Timeout"
            )
        elif not channel_receiver and not w:
            self.close()
            if not self.reinit_client():
                time.sleep(60)
                raise TransportException(
                    "Port d'émission: lors de l'envoi de l'order - Erreur de Timeout"
                )

        totalsent = 0
        msglen = len(data)

        while totalsent < msglen:

            try:
                if channel_receiver:
                    sent = self.transport_recv.send(data[totalsent:])
                    self.logger.info("Acquittement envoyé sur le port de réception")
                else:
                    sent = self.transport_send.send(data[totalsent:])
                    self.logger.info("Order envoyé sur le port d'émission")
            except socket.error as e:
                if e.args[0] in (
                    errno.EBADF,
                    errno.ECONNRESET,
                    errno.ENOTCONN,
                    errno.ESHUTDOWN,
                    errno.ECONNABORTED,
                    errno.ETIMEDOUT,
                ):
                    self.logger.info(
                        "write error channel_receiver:%s totalsent:%s\n%s \n from %s",
                        channel_receiver,
                        totalsent,
                        data[:totalsent],
                        traceback.format_exc(),
                    )
                    self.close()
                    time.sleep(6)  # reprendre rapidement la com
                    raise TransportException("Disconnected (errno=%d)" % (e.args[0],))
                else:
                    self.close()
                    time.sleep(60)
                    raise

            if sent == 0:
                self.close()
                time.sleep(60)
                raise TransportException("Disconnected (sent 0 bytes)")

            totalsent += sent

    def close(self):
        try:
            # Fermer le socket du canal d'émission serveur tcp/ip
            if self.transport_send is not None:
                self.transport_send.shutdown(socket.SHUT_RDWR)
                self.transport_send.close()
                self.transport_send = None
                if self.socket_send is not None:
                    self.socket_send.close()
                    self.socket_send = None
                    self.logger.info("Fermeture du socket du port d'émission")
        except socket.error as e:
            self.logger.info(
                "Erreur, lors de la fermeture du socket du port d'émission: %s", e
            )

        try:
            # Fermer le socket du canal de réception serveur tcp/ip
            if self.socket_recv is not None:
                self.socket_recv.shutdown(socket.SHUT_RDWR)
                self.socket_recv.close()
                self.socket_recv = None
                self.logger.info("Fermeture du socket du port de réception")
        except socket.error as e:
            self.logger.info(
                "Erreur, lors de la fermeture du socket du port de réception: %s", e
            )
        try:
            if self.transport_recv is not None:
                self.logger.info("Fermeture du serveur de réception")
                self.transport_recv.shutdown(socket.SHUT_RDWR)
                self.transport_recv.close()
                self.transport_recv = None
        except socket.error as e:
            self.logger.info(
                "Erreur lors de la fermeture du serveur de réception: %s", e
            )
            if e.args[0] not in (errno.ENOTCONN, errno.EBADF):
                raise
